#!/usr/bin/perl

use Getopt::Std;
use Text::ParseWords;

# Penalty for occurence of a totally unknown two-letter
# hash
my $unknown_hash_penalty = 1.0;
# Penalty for a previously unknown letter following
# a previously known two-letter combination
my $unknown_follower_penalty = 0.5;

# This is a hash of hashes where the frequency for a certain
# letter following a given two-letter combination is stored.
my %freqhash;

print "Hello. I am a word classifying program.\n";
print "I can tell whether words are meaningful or nonsensical.\n";

getopts("dl:o:", \%args);

# Select language
if ($args{l}) {
    $language = lc($args{l});
} else {
    $language = "en";
}
read_frequency_table("stats/frequencies.${language}");

# Dump frequencies
if ($args{d}) {
    dump_hash();
}

foreach $file (@ARGV) {
    classify_words_in_file($file);
}

############################################################################################
# END OF MAIN PROGRAM
############################################################################################

sub classify_words_in_file {
    print "Classifying words in file $_[0]\n";
    open(FD, $_[0])
	or die "can't open $_[0]: $!\n";
    my $wordcount = 0;
    my $nonsensecount = 0;
    while (!eof(FD)) {
	$line = readline(FD);
	$line = lc($line);
	# Regexp to parse strings.
	@words = &parse_line('\s+', 0, $line);
	foreach $word (@words) {
	    $word=~s/[\.\(\)!\?\+,0-9]//g;
	    # Classify words longer than 5 characters
	    if (length($word) > 5) {
		my $status = classify_word($word);
		if (!$status) {
		    $nonsensecount++;
		}
		$wordcount++;
	    }
	}
    }
    print ("$wordcount words, $nonsensecount nonsense words\n");
    print ("That's ", ($nonsensecount/$wordcount)*100, " percent.\n");
    close(FD);

}

sub classify_word {
    print "Classifying word \"$_[0]\"...";
    my @vals = unpack("C*", $_[0]);
    my $chars = $#vals + 1;
    # print " number of chars: ", $chars;
    my $i = 0;
    # Initial probability score.
    my $score = 0.0;
    foreach $ch (@vals) {
	if ($i > 1) {
	    # The two previous chars consitute a hash value
	    $key = @vals[$i-2]*256 + @vals[$i-1];
	    # Then this char is the value at that point
	    $val = @vals[$i];
	    # print ("Hashkey: $key, Value: $val\n");
	    # see if a statistic exists for these two letters
	    if (exists($freqhash{$key})) {
		if (exists($freqhash{$key}{$val})) {
		    # Weight with the number of chars following the first
		    # two chars, in ordet to be fair - long word means
		    # bigger chance of hitting a valid combination and
		    # so it must be weighted.
		    $score = $score + ($freqhash{$key}{$val} / ($chars-2));
		} else {
		    # Penalty for non-encountered character after this hash
		    $score = $score - ($unknown_follower_penalty / ($chars - 2));
		}		
	    } else {
		# Penalty for non-existing hash.
		$score = $score - ($unknown_hash_penalty / ($chars - 2));
	    }
	}
	$i++;
    }
    # Arbitrary set limit - classify very few words as
    # nonsens, but those who are are almost always nonsensical.
    $status = ($score > 0.1);
    if ($status) {
	print(" meaningful ");
    } else {
	print(" nonsense ");
    }
    print "(score: ", $score, ")\n";
    return ($status);
}

sub dump_hash {
    # Iterate through the hash and print out hash value,
    # all chars following it, and their respective occurencies
    # and frequencies.
    print ("Dumping hash...\n");
    foreach $key ( keys %freqhash ) {
	print "$key: {\n";
	foreach $stat ( keys %{ $freqhash{$key} } ) {
	    print ("  ", $stat, " => ", $freqhash{$key}{$stat}, "\n");
	}
	print "}\n";
    }
}

# This function reads in a frequency table for a certain language
# as generated by the "getstats.pl" program.
sub read_frequency_table {
    open(FD, "$_[0]")
	or die "can't open frequency table $_[0]: $!\n";
    while (!eof(FD)) {
	$line = readline(FD);
	@row = &parse_line(':', 0, $line);
	$key = $row[0] + 0;
	@values = &parse_line(',', 0, $row[1]);
	$i = 0;
	foreach $a (@values) {
	    # Remove brackets and stuff (no good parsing...)
	    $a=~s/[^0-9\.]//g;
	    # Convert to figure
	    $a=$a+0;
	    if ($i%2 == 0) {
		# Even, key
		$char = $a;
	    } else {
		$freq = $a;
		# Add to hash
		$freqhash{$key}{$char} = $freq;
	    }
	    $i++;
	}
    }
    close(FD);
}


