#!/usr/bin/perl

use Getopt::Std;
use Text::ParseWords;

# This is a hash of arrays containing (char, occ, freq) triplets
# where char = char following that hash value
#       occ  = number of occurences
#       freq = the calculated frequency for this char following
#              that hash.
my %freqhash;

print "Hello. I am a tokenizing statistics gathering program.\n";

# sets $args{o}.
getopts("do:", \%args);

foreach $file (@ARGV) {
    add_stats_from_file($file);
    calculate_frequencies();
}

# Dump frequencies
if ($args{d}) {
    dump_hash();
}

# Write out statistics to file
if (length($args{o}) > 0) {
    write_frequency_table($args{o});
}

############################################################################################
# END OF MAIN PROGRAM
############################################################################################

sub add_stats_from_file {
    print "Adding stats from $_[0]\n";
    open(FD, $_[0])
	or die "can't open $_[0]: $!\n";
    while (!eof(FD)) {
	$line = readline(FD);
	$line = lc($line);
	# Regexp to parse strings.
	@words = &parse_line('\s+', 0, $line);
	foreach $word (@words) {
	    # $word=~s/[\.\(\)!\?\+,0-9]//g;
	    # Add stats for words longer than 5 characters
	    if (length($word) > 5) {
		add_word($word);
	    }
	}
    }
    close(FD);
}

sub add_word {
    # print "Adding stats for <$_[0]>\n";
    # Why in hell only the first???
    @vals = unpack("C*", $_[0]);
    #print("Word length: ", $#vals+1, "\n");
    $i = 0;
    foreach $ch (@vals) {
	# print ("$i: $ch\n");
	if ($i > 1) {
	    # The two previous chars consitute a hash value
	    $key = @vals[$i-2]*256 + @vals[$i-1];
	    # Then this char is the value at that point
	    $val = @vals[$i];
	    # print ("Hashkey: $key, Value: $val\n");

	    # Increase the number of times this letter has followed
	    # the previous two letters, e.g. add one to the hash
	    # value for this letter.
	    # hashval -> list of letters following this hash, and their
	    #            repsective occurences.
	    
	    #$aref = \@array;         # $aref now holds a reference to @array
	    #reverse @{$aref};        # Using a $aref as an array

	    # see if a statistic exists for these two letters
	    if (exists($freqhash{$key})) {
		if (exists($freqhash{$key}{$val})) {
		    # Increase the frequency.
		    @{ $freqhash{$key}{$val} }[0]++;
		    #print ("Increased stats for char $val to $freqhash{$key}{$val}[0]\n");
		} else {
		    @{ $freqhash{$key}{$val} } = (1, 0.0);
		    # $freqhash{$key} = $charhash;
		    #print("Added first stats for $val following $key\n");
		}
	    } else {
		# else create a new two-letter hash and default count=1
		@{ $freqhash{$key}{$val} } = (1, 0.0);
		#print ("Created new hash entry for $key\n");
		#print ("Added first stats for $val following $key\n");
	    }
	}
	$i++;
    }
}

sub calculate_frequencies {
    # Take the hash, and calculate for each hash point (e.g
    # each two letters combo), the probability that any given 
    # letter will follow these two.
    print ("Calculating frequencies...\n");
    foreach $key ( keys %freqhash ) {
	#print "Calculating stats for $key...\n";
	my $total = 0;
	foreach $stat ( keys %{ $freqhash{$key} } ) {
	    $total = $total + $freqhash{$key}{$stat}[0];
	}
	foreach $stat ( keys %{ $freqhash{$key} } ) {
	    $freqhash{$key}{$stat}[1] = $freqhash{$key}{$stat}[0] / $total;
	}
    }
}

sub dump_hash {
    # Iterate through the hash and print out hash value,
    # all chars following it, and their respective occurencies
    # and frequencies.
    print ("Dumping hash...\n");
    foreach $key ( keys %freqhash ) {
	print "$key: {\n";
	foreach $stat ( keys %{ $freqhash{$key} } ) {
	    print ("  ", $stat, " => ", $freqhash{$key}{$stat}[0], " times");
	    print (", frequency: ", $freqhash{$key}{$stat}[1], "\n");
	}
	print "}\n";
    }
}

sub write_frequency_table {
    # Write out a frequency file: hashvalue (byte-value of
    # first character * 256 + byte-value of second character)
    # followed by a list of possible following letters, and
    # their respective frequencies.
    # Hash -> list of letters
    #         each gives      ->  letter value (byte)
    #                         ->  probability (0-1) (float)
    # If a letter is non-existing this means the probability
    # for this combination is 0 (acronym, new word, or nonsense).
    open(FD, "> $_[0]")
	or die "can't open $_[0]: $!\n";
    foreach $key ( keys %freqhash ) {
	print FD "$key:{";
	$i = 0;
	foreach $stat ( keys %{ $freqhash{$key} } ) {
	    if ($i > 0) {
		print FD ",";
	    }
	    print FD $stat;
	    print FD ",";
	    print FD $freqhash{$key}{$stat}[1];
	    $i++;
	}
	print FD "}\n";
    }
    close(FD);
}
