#!/usr/bin/perl

use Getopt::Std;
use Text::ParseWords;

# Penalty for occurence of a totally unknown two-letter
# hash
my $unknown_hash_penalty = 1.0;
# Penalty for a previously unknown letter following
# a previously known two-letter combination
my $unknown_follower_penalty = 0.5;

# This is a hash of hashes where the frequency for a certain
# letter following a given two-letter combination is stored.
my %freqhash;
my $verbose = 0;

# A list of ISO 639-1 language codes
my %langcodes = ("ad" => "Afrikaans",
		 "ar" => "Arabic",
		 "be" => "Belarusian",
		 "bg" => "Bulgarian",
		 "cs" => "Czech",
		 "da" => "Danish",
		 "de" => "German",
		 "en" => "English",
		 "es" => "Spanish",
		 "fi" => "Finnish",
		 "fr" => "French",
		 "hu" => "Hungarian",
		 "id" => "Indonesian",
		 "is" => "Icelandic",
		 "it" => "Italian",
		 "ja" => "Japanese",
		 "la" => "Latin", 
		 "nl" => "Dutch", 
		 "no" => "Norweigan",
		 "pl" => "Polish",
		 "pt" => "Portuges", 
		 "ru" => "Russian", 
		 "sv" => "Swedish",
		 "tr" => "Turkish",
);

print "Hello. I am a language guesser program.\n";
print "I can sort of tell what language a text is written in.\n";

getopts("vl:o:", \%args);

if ($args{v}) {
    $verbose = 1;
}

foreach $file (@ARGV) {
    test_languages($file);
}

############################################################################################
# END OF MAIN PROGRAM
############################################################################################

sub test_languages {
    # Find available languages
    $languages = 0;
    $tobetested = $_[0];
    opendir(DIR, "stats");
    my @languages;
    my @scores;
    @files = readdir(DIR)
	or die "can't opendir $some_dir: $!\n";
    closedir(DIR);
    foreach $file (@files)
    {
	if (length($file) == 14) {
	    $languages++;
	    $file=~s/([a-z]*)\.([a-z]{2})/\2/;
	    $languages[++$#languages] = $file;
	    if ($verbose) {
		print ("$file:");
	    }
	    $score = get_score_for_language($tobetested, $file);
	    if ($verbose) {
		print ("$score\n");
	    }
	    $scores[++$#scores] = $score;	    
	}
    }
    if ($verbose) {
	print("\n");
    }

    # Find the highest score.
    $i = 0;
    $bestscore = 0.0;
    $bestindex = -1;
    for $score (@scores) {
	if ($score > $bestscore) {
	    $bestscore = $score;
	    $bestindex = $i;
	}
	$i++;
    }
    
    if ($bestindex < 0) {
	print ("I don't recognize this language. Sorry.\n");
    } else {
	if ($verbose) {
	    print ("Best match: @languages[$bestindex]\n");
	}
	my $langcode = @languages[$bestindex];
	my $language = $langcodes{$langcode};
	print ("This looks like $language!\n");
    }
}

sub get_score_for_language {
    # Clear hash for each new language
    foreach $key (keys %freqhash) {
	delete $freqhash{$key};
    }
    # print "Testing $_[0] against language $_[1] ...\n";
    read_frequency_table("stats/frequencies.$_[1]");
    return get_score_for_file($_[0]);
}

sub get_score_for_file {
    # print "Classifying words in file $_[0]\n";
    open(FD, $_[0])
	or die "can't open $_[0]: $!\n";
    my $score = 0;
    while (!eof(FD)) {
	$line = readline(FD);
	$line = lc($line);
	# Regexp to parse strings.
	@words = &parse_line('\s+', 0, $line);
	foreach $word (@words) {
	    $word=~s/[\.\(\)!\?\+,0-9]//g;
	    # Classify words longer than 5 characters
	    if (length($word) > 5) {
		$score = $score + word_score($word);
	    }
	}
    }
    close(FD);
    return $score;
}

sub word_score {
    # print "Getting score for word \"$_[0]\"...";
    my @vals = unpack("C*", $_[0]);
    my $chars = $#vals + 1;
    # print " number of chars: ", $chars;
    my $i = 0;
    # Initial probability score.
    my $score = 0.0;
    foreach $ch (@vals) {
	if ($i > 1) {
	    # The two previous chars consitute a hash value
	    $key = @vals[$i-2]*256 + @vals[$i-1];
	    # Then this char is the value at that point
	    $val = @vals[$i];
	    # print ("Hashkey: $key, Value: $val\n");
	    # see if a statistic exists for these two letters
	    if (exists($freqhash{$key})) {
		if (exists($freqhash{$key}{$val})) {
		    # Weight with the number of chars following the first
		    # two chars, in ordet to be fair - long word means
		    # bigger chance of hitting a valid combination and
		    # so it must be weighted.
		    $score = $score + ($freqhash{$key}{$val} / ($chars-2));
		} else {
		    # Penalty for non-encountered character after this hash
		    $score = $score - ($unknown_follower_penalty / ($chars - 2));
		}		
	    } else {
		# Penalty for non-existing hash.
		$score = $score - ($unknown_hash_penalty / ($chars - 2));
	    }
	}
	$i++;
    }
    return ($score);
}

# This function reads in a frequency table for a certain language
# as generated by the "getstats.pl" program.
sub read_frequency_table {
    open(FD, "$_[0]")
	or die "can't open frequency table $_[0]: $!\n";
    while (!eof(FD)) {
	$line = readline(FD);
	@row = &parse_line(':', 0, $line);
	$key = $row[0] + 0;
	@values = &parse_line(',', 0, $row[1]);
	$i = 0;
	foreach $a (@values) {
	    # Remove brackets and stuff (no good parsing...)
	    $a=~s/[^0-9\.]//g;
	    # Convert to figure
	    $a=$a+0;
	    if ($i%2 == 0) {
		# Even, key
		$char = $a;
	    } else {
		$freq = $a;
		# Add to hash
		$freqhash{$key}{$char} = $freq;
	    }
	    $i++;
	}
    }
    close(FD);
    return $freqhash;
}


